/*
** Module   : ButtonGuiImpl.java
** Abstract : GUI push button implementation.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20141120 Initial version.
** 002 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.  Do not draw directly from APIs which are   
**                  called from mouse events: instead, rely on repainting, which posts proper event
**                  to the event chain and process the drawing in the parent's context.
** 003 CA  20141206 Improvement of H002 - refresh needs to call repaint() always.
** 004 EVL 20141211 Detaching tooltip code from button widget.  Adding image loading for button.
** 005 EVL 20141226 Using images to draw button implementation.
** 006 EVL 20150115 Rework for image usage implementation.
** 007 EVL 20150121 Fixup image coordinates while loading.  Adding handling of the deferred
**                  images.
** 008 EVL 20150210 Fix for button background while image drawing.
** 009 EVL 20150220 Merge CONVERT-3D-COLORS flag to image widgets used to draw the button.
** 010 EVL 20150224 NO-CONVERT-3D-COLORS option is now processing with CONVERT-3D-COLORS attribute
**                  that is really the same(with boolean inversion).
** 011 EVL 20150313 Adding support for NO-FOCUS, FLAT-BUTTON attributes. Default button can not
**                  display any images, only size adjust.
** 012 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 013 EVL 20150320 Fix for GO event processing.  We need to unlock the update key reader.  Fix
**                  for mouse remains pressed even pointer is leaving the button.
** 014 EVL 20150413 Fixing clipping issues.
** 015 HC  20150323 Removed business logic from config classes.
** 016 EVL 20150428 Adding support for prepackaged images.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of RECTANGLE widget.
 */
public class ButtonGuiImpl
extends Button<GuiOutputManager>
{
   /** Prebuilt image size of the arrow style button. */
   public static final int PREBUILT_SIZE = 15;
   
   /** Dotted line width. */
   public static final float DOT_LINE_WIDTH = 1.0f;
   
   /** Shift in XY to draw images on the button. */
   private static final int IMAGE_SHIFT = 3;
   
   /** Dotted line specific parameter. */
   private static final float DOT_STEP = 2.0f;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Light shadow box color. */
   private Color lightShadow;

   /** Dark shadow box color. */
   private Color darkShadow;
   
   /** Very dark shadow box color. */
   private Color veryDarkShadow;
   
   /** Dotted line pattern for focused button text frame. */
   private Stroke dotStroke;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Image to draw on button in up state. */
   private ImageGuiImpl imgUp = null; 
   
   /** Image to draw on button in pressed state. */
   private ImageGuiImpl imgDown = null; 
   
   /** Image to draw on button in disabled state. */
   private ImageGuiImpl imgDisabled = null;
   
   /** True when mouse is over the flat button.  */
   private boolean mouseEntered = false;
   
   /**
    * This constructor creates a button with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   ButtonGuiImpl(WidgetId id, Container<GuiOutputManager> owner, String text) 
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ButtonGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ButtonConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ButtonConfig wcfg)
   {
      super.initialize(id, wcfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // output manager
      screen = screen();
      cc = screen.coordinates();
      // font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, config(), false);
      // initialize the shadow colors
      lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      // initialize dot pattern 
      dotStroke = new BasicStroke(DOT_LINE_WIDTH, BasicStroke.CAP_BUTT,
                                  BasicStroke.JOIN_BEVEL, 0.0f, // this must be just valid
                                  new float[] {DOT_STEP, DOT_STEP}, DOT_STEP);

      // register tooltip action listener
      gd.registerTooltipWorker(this, gf.font().font, ToolTip.isActive(config().tooltip));
      // handle pending images to preload
      loadPendingImages();
      // adjust image coordinates and width
      recalcImagesCoords();
   }
   
   /**
    * Sets the new image for given button widget on the client side.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    image
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(int type, byte[] image)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            if (imgUp == null)
            {
               imgUp = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgUp.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp, config().imgUpDef.name != null ? config().imgUpDef
                                                                     : config().imgDef);
               // prepackaged images has transparent background
               String imgName = config().imgUpDef.name != null
                                   ? config().imgUpDef.name : config().imgDef.name;
               if (ImageHelper.isPrepackaged(imgName))
               {
                  ((ImageConfig)imgUp.config()).transparent = true;
               }
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               imgDown = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDown.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown, config().imgDownDef);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               imgDisabled = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDisabled.setImage(image);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled, config().imgInsDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(config().imgInsDef.name))
               {
                  ((ImageConfig)imgDisabled.config()).transparent = true;
               }
            }
            break;
         }
      }
      
      return result;
   }
   
   /**
    * Loads the new image for given button widget from the client side filesystem.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(int type, String fullName)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            if (imgUp == null)
            {
               imgUp = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgUp.loadImage(fullName);
            
            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp, config().imgUpDef.name != null ? config().imgUpDef
                                                                     : config().imgDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(fullName))
               {
                  ((ImageConfig)imgUp.config()).transparent = true;
               }
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               imgDown = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDown.loadImage(fullName); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown, config().imgDownDef);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               imgDisabled = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDisabled.loadImage(fullName); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled, config().imgInsDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(fullName))
               {
                  ((ImageConfig)imgDisabled.config()).transparent = true;
               }
            }
            break;
         }
      }
    
      // if the image was loaded after button initialization, we need to recalculate new
      // position inside the button
      if (result)
      {
         recalcImagesCoords();
      }
      
      return result;
   }
   
   /**
    * In GUI mode, the drawing needs to be delayed.
    */
   @Override
   protected void refresh()
   {
      repaint();
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      // button options need to be refreshed before every painting
      final ButtonConfig bcfg = config();
      gc = new GuiColorResolver(gd, bcfg, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final String text = textLabel();
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = bcfg.noFocus && imgDown != null
         ? clipRectangle(p, new NativeDimension(d.width, d.height))
         : clipRectangle(p, new NativeDimension(d.width + 1, d.height + 1));
      
      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            // button state flags
            boolean highlighted = highlighted();
            
            // flat button flag
            boolean flat = bcfg.noFocus && bcfg.flatButton;

            // having prepacked images
            boolean builtInImages = false; 

            // draw button itself
            if (hasImages() && !isDefaultButton())
            {
               // check for built-in imaged
               builtInImages = hasPrepackagedImages(bcfg);
               
               // button border is included in methods below
               if (builtInImages)
               {
                  drawPrepackagedImages(po, d, highlighted, pressed, flat);
               }
               else
               {
                  drawButtonImages(po, d, highlighted, pressed, flat);
               }
            }
            else
            {
               // draw button border
               drawButtonBorder(po, d, highlighted, pressed, flat);
               // abd the body
               drawButtonInternal(po, d, text, fntCache.font, highlighted, pressed, flat);
            }
            
            // focused button needs special highlight
            if (highlighted)
            {
               drawHighlight(po, d, gc.fgColor, builtInImages);
            }
         }
         
      });
   }

   /**
    * Checks if this button works as default one.  The frame DEFAULT-BUTTON attribute should also
    * be considered, not only DEFAULT attribute of the button itself.
    *
    * @return   <code>TRUE</code> if button works as default one <code>FALSE</code> otherwise.
    */
   private boolean isDefaultButton()
   {
      // look frame contains this button
      Frame tfr = (Frame) UiUtils.locateFrame(this);
      // get the current button config
      ButtonConfig bcfg = config();
      
      return bcfg.isDefault && tfr != null && tfr.config().defaultButtonId == bcfg.id.asInt();
   }
   
   /**
    * Sets up initial image location inside the button widget.
    * 
    * @param   imgToFix
    *          Image object to fix the coordinates.
    * @param   def
    *          Image definition object to override the image under fixup.
    */
   private void fixupImageCoord(ImageGuiImpl imgToFix, ButtonImageDefinition def)
   {
      // image position
      if (config().noFocus && imgDown == null)
      {
         imgToFix.setPhysicalLocation(IMAGE_SHIFT, IMAGE_SHIFT);
      }
      // getting reference to image config
      ImageConfig icfg = (ImageConfig) imgToFix.config();
      // apply new data but only if we changed defaults
      // offsets
      if (def.offsetX != -1)
      {
         icfg.offsetX = def.offsetX;
      }
      else if (def.offsetCol != 0.0)
      {
         icfg.offsetCol = def.offsetCol;
      }
      if (def.offsetY != -1)
      {
         icfg.offsetY = def.offsetY;
      }
      else if (def.offsetRow != 0.0)
      {
         icfg.offsetRow = def.offsetRow;
      }
      // and size too
      if (def.widthPixels != 0)
      {
         icfg.widthPixels = def.widthPixels;
      }
      else if (def.widthChars != 0.0)
      {
         icfg.widthChars = def.widthChars;
      }
      if (def.heightPixels != 0)
      {
         icfg.heightPixels = def.heightPixels;
      }
      else if (def.heightChars != 0.0)
      {
         icfg.heightChars  = def.heightChars;
      }
      // finally merge the convert 3d colors flags for used image
      icfg.convert3D = config().convert3D;
   }
   
   /**
    * Draw highlight frame indicating focused button.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   color
    *          Border color.
    * @param   prepack
    *          Flag indicating the button has prepackaged images.
    */
   private void drawHighlight(NativePoint orig,
                              NativeDimension dim,
                              Color color,
                              boolean prepack)
   {
      gd.setColor(color);
      if (imgDown != null || prepack)
      {
         gd.drawRect(orig.x, orig.y, dim.width - 1, dim.height - 1);
      }
      else if (imgUp != null)
      {
         gd.drawRect(orig.x, orig.y, dim.width + 1, dim.height + 1);
      }
      else
      {
         gd.drawRect(orig.x, orig.y, dim.width, dim.height);
      }
   }
   
   /**
    * Draw bottom border simulating 3D effect. 
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          New flat button style flag.
    */
   private void drawButtonBorder(NativePoint orig,
                                 NativeDimension dim,
                                 boolean focused,
                                 boolean pressed,
                                 boolean flat)
   {
      // coordinates are dependent from button state
      int x1 = focused ? orig.x + 1 : orig.x;
      int x2 = focused ? orig.x + dim.width - 1 : orig.x + dim.width;
      int y1 = focused ? orig.y + 1 : orig.y;
      int y2 = focused ? orig.y + dim.height - 1 : orig.y + dim.height;

      if (pressed)
      {
         if (flat)
         {
            gd.setColor(darkShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);

            gd.setColor(lightShadow);
            gd.drawLine(x1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1);
         }
         else
         {
            gd.setColor(darkShadow);
            gd.drawRect(x1, y1, x2 - x1, y2 - y1);
         }
      }
      else
      {
         if (flat && mouseEntered)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);

            gd.setColor(darkShadow);
            gd.drawLine(x1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1);
         }
         else if (!flat)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);

            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 1, x2 - 1, y1 + 1);

            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1);
         }
      }
   }
   
   /**
    * Draw button internals with different states.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   label
    *          Button label.
    * @param   fnt
    *          Font to use for button label.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          New flat button style flag.
    */
   private void drawButtonInternal(NativePoint orig,
                                   NativeDimension dim,
                                   String label,
                                   Font fnt,
                                   boolean focused,
                                   boolean pressed,
                                   boolean flat)
   {
      // coordinates are dependent from button state
      int x = focused ? orig.x + 1 : orig.x;
      int y = focused ? orig.y + 1 : orig.y;
      int w = focused ? dim.width - 2 : dim.width;
      int h = focused ? dim.height - 2 : dim.height;

      // draw button body
      gd.setColor(gc.bgColor);
      if (pressed)
      {
         // flat and non-flat modes behave the same way here
         gd.fillRect(x + 1, y + 1, w - 1, h - 1);
      }
      else
      {
         if (flat)
         {
            if (mouseEntered)
            {
               gd.fillRect(x + 1, y + 1, w - 1, h - 1);
            }
            else
            {
               gd.fillRect(x, y, w + 1, h + 1);
            }
         }
         else
         {
            gd.fillRect(x + 1, y + 1, w - 2, h - 2);
         }
      }
      
      // draw button text
      gd.setGuiFont(fnt);
      int xt = orig.x + (dim.width - gd.getTextWidth(label, fnt))/2;
      int yt = orig.y + (dim.height + fnt.getSize())/2;
      gd.setFontStyle(fnt.getStyle());
      // enabled state controls text drawing
      if (isEnabled())
      {
         gd.setColor(gc.fgColor);
         if (pressed)
         {
            gd.drawString(label, xt + 1, yt + 1);
         }
         else
         {
            gd.drawString(label, xt, yt);
         }
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(label, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(label, xt, yt);
      }

      // draw dotted rectangle around focused button text
      if (isEnabled() && focused)
      {
         // set dotted line style
         gd.setLineStroke(dotStroke);
         int dx = 2;
         int dy = 2;
         if (pressed)
         {
            gd.drawRect(x + dx + 2, y + dy + 2, w - 2*dx - 2, h - 2*dy - 2);
         }
         else
         {
            gd.drawRect(x + dx + 1, y + dy + 1, w - 2*dx - 2, h - 2*dy - 2);
         }
         // restore default line style
         gd.setLineStroke(new BasicStroke());
      }
   }
   
   /**
    * Draw button internals with different states using preloaded images.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          New flat button style flag.
    */
   private void drawButtonImages(NativePoint orig,
                                 NativeDimension dim,
                                 boolean focused,
                                 boolean pressed,
                                 boolean flat)
   {
      // NO-FOCUS attribute flag value
      boolean noFocusFlag = config().noFocus; 
      // coordinates are dependent from button state
      int x1 = noFocusFlag ? orig.x : orig.x + 1;
      int x2 = orig.x + dim.width - 1;
      int y1 = noFocusFlag ? orig.y : orig.y + 1;
      int y2 = orig.y + dim.height - 1;

      // draw the light frame around image
      gd.setColor(gc.bgColor);
      if (noFocusFlag && imgDown == null)
      {
         gd.fillRect(x1, y1, x2 - x1 + 2, y2 - y1 + 2);
      }
      else if (!noFocusFlag && imgDown != null)
      {
         gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      }
      else if (!noFocusFlag && imgDown == null)
      {
         gd.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
      }

      // draw button body images
      if (pressed)
      {
         // determine what image to draw
         ImageGuiImpl imgPressed = null;
         if (imgDown != null)
         {
            imgPressed = imgDown; 
         }
         else if (imgUp != null)
         {
            imgPressed = imgUp; 
         }
         // draw pressed state
         if (imgPressed != null)
         {
            // for pressed state need 1 pixel right-down shift 
            NativePoint location = imgPressed.physicalLocation();
            if (imgDown == null)
            {
               imgPressed.setPhysicalLocation(location.x + 1, location.y + 1);
            }
            imgPressed.draw();
            // and restore origin after draw
            imgPressed.setPhysicalLocation(location.x, location.y);
         }
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (isEnabled() && imgUp != null)
         {
            imgReleased = imgUp; 
         }
         else if (!isEnabled() && imgDisabled != null)
         {
            imgReleased = imgDisabled; 
         }
         // draw released state
         if (imgReleased != null)
         {
            imgReleased.draw();
         }
         else
         {
            // draw the dark filled rectangle instead of image
            gd.setColor(darkShadow);
            gd.fillRect(x1 + 2, y1 + 2, x2 - x1 - 3, y2 - y1 - 3);
            // and lines marks the image as disabled
            gd.setColor(lightShadow);
            gd.drawLine(x1 + 3, y2 - 1, x2 - 2, y2 - 1);
            gd.drawLine(x2 - 1, y1 + 3, x2 - 1, y2 - 1);
         }
      }

      // draw 3D shadow for image button
      if (pressed)
      {
         if (flat)
         {
            if (imgDown == null)
            {
               gd.setColor(darkShadow);
               gd.drawLine(x1, y1, x1, y2);
               gd.drawLine(x1, y1, x2, y1);

               gd.setColor(lightShadow);
               gd.drawLine(x1, y2 + 1, x2, y2 + 1);
               gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);
            }
            else
            {
               gd.setColor(darkShadow);
               gd.drawLine(x1, y1, x1, y2 - 1);
               gd.drawLine(x1, y1, x2 - 1, y1);

               gd.setColor(lightShadow);
               gd.drawLine(x1, y2, x2 - 1, y2);
               gd.drawLine(x2, y2, x2, y1);
            }
         }
         else if (imgDown == null)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y2 + 1, x2, y2 + 1);
            gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);

            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2, x1 + 1, y1 + 1);
            gd.drawLine(x1 + 1, y1 + 1, x2, y1 + 1);

            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y1, x1, y2);
            gd.drawLine(x1, y1, x2, y1);
         }
      }
      else
      {
         if (flat && mouseEntered && isEnabled())
         {
            if (imgDown == null)
            {
               gd.setColor(lightShadow);
               gd.drawLine(x1, y1, x1, y2);
               gd.drawLine(x1, y1, x2, y1);

               gd.setColor(darkShadow);
               gd.drawLine(x1, y2 + 1, x2 + 1, y2 + 1);
               gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);
            }
            else
            {
               gd.setColor(lightShadow);
               gd.drawLine(x1, y1, x1, y2 - 1);
               gd.drawLine(x1, y1, x2 - 1, y1);

               gd.setColor(darkShadow);
               gd.drawLine(x1, y2, x2 - 1, y2);
               gd.drawLine(x2, y2, x2, y1);
            }
         }
         else if (!flat && imgDown == null)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2);
            gd.drawLine(x1, y1, x2, y1);


            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1 + 1);

            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y2 + 1, x2 + 1, y2 + 1);
            gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);
         }
      }
   }

   /**
    * Draw button internals with different states using preloaded prepackaged images.
    * 
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          New flat button style flag.
    */
   private void drawPrepackagedImages(NativePoint orig,
                                      NativeDimension dim,
                                      boolean focused,
                                      boolean pressed,
                                      boolean flat)
   {
      // NO-FOCUS attribute flag value
      boolean noFocusFlag = config().noFocus; 
      // coordinates are dependent from button state
      int x1 = noFocusFlag ? orig.x : orig.x + 1;
      int x2 = noFocusFlag ? orig.x + dim.width : orig.x + dim.width - 1;
      int y1 = noFocusFlag ? orig.y : orig.y + 1;
      int y2 = noFocusFlag ? orig.y + dim.height : orig.y + dim.height - 1;

      // draw the light frame around image
      gd.setColor(gc.bgColor);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      
      // draw button body images
      if (pressed)
      {
         // determine what image to draw
         ImageGuiImpl imgPressed = null;
         if (imgDown != null)
         {
            imgPressed = imgDown; 
         }
         else if (imgUp != null)
         {
            imgPressed = imgUp; 
         }
         // draw pressed state
         if (imgPressed != null)
         {
            // for pressed state need 1 pixel right-down shift 
            NativePoint location = imgPressed.physicalLocation();
            imgPressed.setPhysicalLocation(location.x + 1, location.y + 1);
            imgPressed.draw();
            // draw small button frame
            gd.setColor(darkShadow);
            gd.drawRect(location.x, location.y, PREBUILT_SIZE, PREBUILT_SIZE);
            // and restore origin after draw
            imgPressed.setPhysicalLocation(location.x, location.y);
         }
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (!isEnabled() && imgDisabled != null)
         {
            imgReleased = imgDisabled; 
         }
         else if (imgUp != null)
         {
            imgReleased = imgUp; 
         }
         // draw released state
         if (imgReleased != null)
         {
            imgReleased.draw();
            // draw small button frame
            NativePoint location = imgReleased.physicalLocation();
            gd.setColor(lightShadow);
            gd.drawLine(location.x + 1, location.y + 1,
                        location.x + 1, location.y + PREBUILT_SIZE - 2);
            gd.drawLine(location.x + 1, location.y + 1,
                        location.x + PREBUILT_SIZE - 2, location.y + 1);
            gd.setColor(darkShadow);
            gd.drawLine(location.x + 1, location.y + PREBUILT_SIZE - 1,
                        location.x + PREBUILT_SIZE - 1, location.y + PREBUILT_SIZE - 1);
            gd.drawLine(location.x + PREBUILT_SIZE - 1, location.y + 1,
                        location.x + PREBUILT_SIZE - 1, location.y + PREBUILT_SIZE - 2);
            gd.setColor(veryDarkShadow);
            gd.drawLine(location.x, location.y + PREBUILT_SIZE,
                        location.x + PREBUILT_SIZE, location.y + PREBUILT_SIZE);
            gd.drawLine(location.x + PREBUILT_SIZE, location.y,
                        location.x + PREBUILT_SIZE, location.y + PREBUILT_SIZE - 1);
         }
      }

      // draw 3D shadow for image button
      if (pressed)
      {
         if (flat)
         {
            gd.setColor(darkShadow);
            gd.drawLine(x1, y1, x1, y2 - 2);
            gd.drawLine(x1 + 1, y1, x2 - 2, y1);

            gd.setColor(lightShadow);
            gd.drawLine(x1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 2, x2 - 1, y1);
         }
      }
      else
      {
         if (flat && mouseEntered && isEnabled())
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 2);
            gd.drawLine(x1, y1, x2 - 2, y1);

            gd.setColor(darkShadow);
            gd.drawLine(x1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 2, x2 - 1, y1);
         }
      }
   }

   /**
    * Check if the button has any images to be used to draw button.
    * 
    * @return  <code>true</code> there is at least one image to use.
    */
   private boolean hasImages()
   {
      return imgUp != null || imgDown != null || imgDisabled != null;
   }

   /**
    * Check if the button has prepackaged images to be used to draw button.
    * 
    * @return  <code>true</code> there is at least one image to use.
    */
   private boolean hasPrepackagedImages(ButtonConfig bcfg)
   {
      return (bcfg.imgDef.name != null && ImageHelper.isPrepackaged(bcfg.imgDef.name)) ||
             (bcfg.imgUpDef.name != null && ImageHelper.isPrepackaged(bcfg.imgUpDef.name));
   }

   /**
    * Need to recalculate image coordinates after all images have been loaded.
    */
   private void recalcImagesCoords()
   {
      // getting button width and height in pixels
      int bw = screen().coordinates().columnToNative(width());
      int bh = screen().coordinates().rowToNative(height());
      if (imgUp != null)
      {
         recalcImageCoord(imgUp, bw, bh);
      }
      if (imgDown != null)
      {
         recalcImageCoord(imgDown, bw, bh);
      }
      if (imgDisabled != null)
      {
         recalcImageCoord(imgDisabled, bw, bh);
      }
   }
   
   /**
    * Reposition image and change the width according button width and height.
    *
    * @param   imgToFix
    *          Image object to fix the coordinates.
    * @param   bw
    *          Button width.
    * @param   bh
    *          Button height.
    */
   private void recalcImageCoord(ImageGuiImpl imgToFix, int bw, int bh)
   {
      // getting image width and height in pixels
      int iw = imgToFix.getImageWidth();
      int ih = imgToFix.getImageHeight();
      NativePoint loc = imgToFix.physicalLocation();
      if (bw > iw)
      {
         loc.x = imgDown != null || config().noFocus || hasPrepackagedImages(config())
                    ? (bw - iw)/2 : 1 + (bw - iw)/2;
      }
      if (bh > ih)
      {
         loc.y = imgDown != null || config().noFocus || hasPrepackagedImages(config())
                    ? (bh - ih)/2 : 1 + (bh - ih)/2;
      }
      imgToFix.setPhysicalLocation(loc.x, loc.y);
   }
   
   /**
    * Refresh the image objects to be used to draw button in different states with images that was
    * saved before in pending storage area.
    */
   private void loadPendingImages()
   {
      String imgName = null;
      // down image first
      imgName = config().imgDownDef.name;
      if (imgName != null && imgName.length() > 0 && imgDown == null)
      {
         loadPendingImage(ButtonConfig.IMAGE_DOWN);
      }
      // up image next
      imgName = (config().imgUpDef.name != null) ? config().imgUpDef.name : config().imgDef.name;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_UP);
      }
      // image for disabled state
      imgName = config().imgInsDef.name;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_INS);
      }
   }

   /**
    * Refresh the image object to be used to draw button in different states with images that was
    * saved before in pending storage area.
    * 
    * @param   type
    *          Image type to check.  Up, Down or Insensitive.
    */
   private void loadPendingImage(int type)
   {
      int thisId = config().id.asInt();

      // first try to search for images loaded from application jar
      byte[] imgData = ImageHelper.getPendingImageData(thisId, type);
      if (imgData != null)
      {
         setImage(type, imgData);
      }
      else
      {
         String loadName = ImageHelper.getPendingImageName(thisId, type);
         if (loadName != null && loadName.length() > 0)
         {
            loadImage(type, loadName);
         }
      }
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      // TODO: this is not correct just to verify mouse handling methods 
      return new NativeRectangle(cc.pointToNative(location()),
                                 cc.dimensionToNative(dimension()));
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      // get text mode height
      double hTxt = super.height();
      // have no images use regular calculation
      if (!hasImages())
      {
         return hTxt;
      }
      
      // compute existed images heights
      double hUp   = imgUp != null ? imgUp.height() : 0.0;
      double hDown = imgDown != null ? imgDown.height() : 0.0;
      double hDis  = imgDisabled != null ? imgDisabled.height() : 0.0;
      
      // when the text mode height defined only by button text we have to ignore this value
      // from image based button height calculation
      if (hTxt <= 1.0)
      {
         // this excludes text base height from Max() computing chain below
         hTxt = 0.0;
      }
      
      // and return max value of them plus image insets
      return config().noFocus && imgDown != null
                       ? Math.max(super.height(), Math.max(hUp, Math.max(hDown, hDis)))
                       : !config().noFocus && imgUp != null && imgDown == null && imgDisabled == null
                              ? Math.max(hTxt, Math.max(hUp, Math.max(hDown, hDis)) +
                                     screen().coordinates().heightFromNative(2*IMAGE_SHIFT - 1))
                              : Math.max(hTxt, Math.max(hUp, Math.max(hDown, hDis)) +
                                     screen().coordinates().heightFromNative(2*IMAGE_SHIFT - 2));
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      // get text mode width
      double wTxt = super.width();
      // have no images use regular calculation
      if (!hasImages())
      {
         return wTxt;
      }
      
      // compute existed images widths
      double wUp   = imgUp != null ? imgUp.width() : 0.0;
      double wDown = imgDown != null ? imgDown.width() : 0.0;
      double wDis  = imgDisabled != null ? imgDisabled.width() : 0.0;
      
      // when the text mode width defined only by button text we have to ignore this value
      // from image based button width calculation
      String lbl = textLabel();
      double wLabel = lbl != null ? (double)lbl.length() + 2.0 : 0.0;
      if (wTxt == wLabel)
      {
         // this excludes text base width from Max() computing chain below
         wTxt = 0.0;
      }
      
      // and return max value of them plus image insets
      return config().noFocus && imgDown != null
                       ? Math.max(super.width(), Math.max(wUp, Math.max(wDown, wDis)))
                       : !config().noFocus && imgUp != null && imgDown == null && imgDisabled == null 
                              ? Math.max(wTxt, Math.max(wUp, Math.max(wDown, wDis)) +
                                  screen().coordinates().widthFromNative(2*IMAGE_SHIFT - 1))
                              : Math.max(wTxt, Math.max(wUp, Math.max(wDown, wDis)) +
                                  screen().coordinates().widthFromNative(2*IMAGE_SHIFT - 2));
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // simulate enter key
      if (isEnabled())
      {
         // generate CHOOSE on button click
         KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_CHOOSE);
         EventManager.postEvent(ki);
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw button in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         requestFocus();
         pressed = true;
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw button in released
    * state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      ButtonConfig bcfg = (ButtonConfig) config();
      
      if (bcfg.noFocus && bcfg.flatButton && !mouseEntered)
      {
         mouseEntered = true;
      }
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (mouseEntered)
      {
         mouseEntered = false;
      }
      if (pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // get coords of the drag event related to button
      NativePoint mp = translate(this, e);
      // get button location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      // this can be rare case when we leave the button with holded mouse key, the button must
      // be released, the mouseExited event will not be delivered to button until drag button
      // released even if the pointer is far away from button accepted press event
      if ((mp.x < p.x || mp.x > (p.x + d.width) || mp.y < p.y || mp.y > (p.y + d.height)) &&
          pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      if (!config().noFocus)
      {
         super.requestFocus();
      }
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return super.focusTraversable() && !config().noFocus;
   }

   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   protected void handleBasicEvents(KeyInput event)
   {
      // special processing for spacebar key in GUI mode
      int id = event.id();
      boolean needRefresh = false;
      
      if (isEnabled())
      {
         // any typed key will upress the button
         if (pressed && id == EventType.KEY_PRESSED)
         {
            pressed = false;
            // state changed, need repainting
            needRefresh = true;
         }
         else
         {
            // space bar is VK_ENTER with KEY_TYPED id
            if (event.keyCode() == Key.VK_ENTER && id == EventType.KEY_TYPED)
            {
               pressed = true;
               // state changed, need repainting
               needRefresh = true;
               // simulate enter key
               EventManager.postEvent(new ActionEvent(this, getCommand()));
            }
         }
         // redraw the button state
         if (needRefresh)
         {
            repaint();
         }
      }
      
      // all other should be processed as usual
      super.handleBasicEvents(event);
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   protected boolean handleCursorKeys(int key, Frame<GuiOutputManager> frame)
   {
      // cursor movement does not work as focus change for GUI button
      if (key != Keyboard.KA_CURSOR_LEFT &&
          key != Keyboard.KA_CURSOR_UP &&
          key != Keyboard.KA_CURSOR_RIGHT &&
          key != Keyboard.KA_CURSOR_DOWN)
      {
         return super.handleCursorKeys(key, frame);
      }
      else
      {
         return false;
      }
   }
}
